import asyncio
import json
import os
import pytest
from nanasqlite_server.client import RemoteNanaSQLite
from nanasqlite_server.cert_gen import generate_certificate
import subprocess
import sys
import time

@pytest.fixture
def test_keys():
    from cryptography.hazmat.primitives.asymmetric import ed25519
    private_key = ed25519.Ed25519PrivateKey.generate()
    public_key = private_key.public_key()
    from cryptography.hazmat.primitives import serialization
    pub_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    ).decode()
    return private_key, pub_bytes

@pytest.fixture
async def multi_db_server(tmp_path, test_keys):
    priv, pub = test_keys
    db_dir = tmp_path / "dbs"
    db_dir.mkdir()
    
    config_path = tmp_path / "accounts.json"
    with open(config_path, "w") as f:
        json.dump({
            "db_dir": str(db_dir),
            "accounts": [
                {
                    "name": "tester",
                    "public_key": pub,
                    "allowed_dbs": ["db1.sqlite", "db2.sqlite", "subdir/db3.sqlite"]
                }
            ]
        }, f)
    
    # Create subdir
    (db_dir / "subdir").mkdir()

    orig_cwd = os.getcwd()
    os.chdir(tmp_path)
    # Start server process
    proc = None
    try:
        generate_certificate()
        
        # Use a random port
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        s.close()

        env = os.environ.copy()
        env["PYTHONPATH"] = orig_cwd + os.pathsep + os.path.join(orig_cwd, "src")
        env["NANASQLITE_FORCE_POLLING"] = "1"
        
        cmd = [
            sys.executable, "-m", "nanasqlite_server.server",
            "--port", str(port),
            "--accounts", str(config_path)
        ]
        
        if sys.platform == "win32":
            creationflags = subprocess.CREATE_NEW_PROCESS_GROUP
        else:
            creationflags = 0

        log_file_path = tmp_path / "multi_db_server.log"
        log_file = open(log_file_path, "w", encoding="utf-8")

        proc = subprocess.Popen(
            cmd, env=env, stdout=log_file, stderr=subprocess.STDOUT, text=True,
            creationflags=creationflags
        )

        # Wait for server to start
        start = time.time()
        ready = False
        
        while time.time() - start < 60:
            if proc.poll() is not None:
                # Process died
                break
            
            # Try to connect to check if server is ready
            try:
                test_client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
                test_client.private_key = priv
                await test_client.connect(account_name="tester")
                await test_client.close()
                ready = True
                break
            except Exception:
                pass
            
            await asyncio.sleep(1.0)
        
        if not ready:
            if proc.poll() is not None:
                log_file.close()
                with open(log_file_path, "r", encoding="utf-8") as f:
                    log_content = f.read()
                raise RuntimeError(f"Server died with code {proc.returncode}. Log:\n{log_content}")
            proc.kill()
            raise RuntimeError("Server failed to start within timeout.")
            
        yield port, priv, db_dir
    finally:
        if proc and proc.poll() is None:
            proc.terminate()
            proc.wait()
        log_file.close()
        os.chdir(orig_cwd)

@pytest.mark.asyncio
async def test_multi_db_access(multi_db_server):
    port, priv, db_dir = multi_db_server
    
    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    
    try:
        await client.connect(account_name="tester")
        
        # Access DB1
        await client.__getattr__("__setitem__")("key1", "val1", db="db1.sqlite")
        assert await client.__getattr__("__getitem__")("key1", db="db1.sqlite") == "val1"
        
        # Access DB2
        await client.__getattr__("__setitem__")("key2", "val2", db="db2.sqlite")
        assert await client.__getattr__("__getitem__")("key2", db="db2.sqlite") == "val2"
        
        # Verify they are separate files
        assert os.path.exists(db_dir / "db1.sqlite")
        assert os.path.exists(db_dir / "db2.sqlite")
        
        # Access Subdir DB
        await client.__getattr__("__setitem__")("key3", "val3", db="subdir/db3.sqlite")
        assert await client.__getattr__("__getitem__")("key3", db="subdir/db3.sqlite") == "val3"
        assert os.path.exists(db_dir / "subdir" / "db3.sqlite")
        
        # Unauthorized DB access
        with pytest.raises(Exception) as exc:
            await client.__getattr__("__getitem__")("key", db="secret.sqlite")
        assert "not allowed" in str(exc.value).lower()
        
        # Traversal attempt
        with pytest.raises(Exception) as exc:
            await client.__getattr__("__getitem__")("key", db="../config.json")
        # Server returns "not allowed" for unauthorized database access
        assert "not allowed" in str(exc.value).lower()

    finally:
        await client.close()



"""
protocol.py の暗号化/復号ユニットテスト

AES-256-GCM のメッセージ暗号化・復号、HKDF セッション鍵導出のテスト。
これらのテストは liboqs-python なしで実行できます。
"""

import os

from nanasqlite_server import protocol


class TestDeriveSessionKey:
    """HKDF セッション鍵導出のテスト"""

    def test_derive_session_key_length(self):
        """セッション鍵が 32 バイト (256 ビット) であることを確認"""
        key = protocol.derive_session_key(b"shared_secret")
        assert len(key) == 32

    def test_derive_session_key_deterministic(self):
        """同じ入力から常に同じ鍵が導出されることを確認"""
        key1 = protocol.derive_session_key(b"same_secret")
        key2 = protocol.derive_session_key(b"same_secret")
        assert key1 == key2

    def test_derive_session_key_different_inputs(self):
        """異なる共有秘密から異なる鍵が導出されることを確認"""
        key1 = protocol.derive_session_key(b"secret_a")
        key2 = protocol.derive_session_key(b"secret_b")
        assert key1 != key2

    def test_derive_session_key_produces_bytes(self):
        """セッション鍵がバイト列であることを確認"""
        key = protocol.derive_session_key(os.urandom(32))
        assert isinstance(key, bytes)


class TestEncryptDecryptMessage:
    """AES-256-GCM 暗号化/復号のテスト"""

    def test_encrypt_decrypt_roundtrip_dict(self):
        """辞書データの暗号化→復号が一致することを確認"""
        session_key = os.urandom(32)
        data = {"method": "get_item_async", "args": ["key"], "kwargs": {}}
        encrypted = protocol.encrypt_message(data, session_key)
        decrypted, rest = protocol.decrypt_message(encrypted, session_key)
        assert decrypted == data
        assert rest == b""

    def test_encrypt_decrypt_roundtrip_string(self):
        """文字列の暗号化→復号が一致することを確認"""
        session_key = os.urandom(32)
        encrypted = protocol.encrypt_message("AUTH_OK", session_key)
        decrypted, rest = protocol.decrypt_message(encrypted, session_key)
        assert decrypted == "AUTH_OK"
        assert rest == b""

    def test_encrypt_decrypt_roundtrip_none(self):
        """None 値の暗号化→復号が一致することを確認"""
        session_key = os.urandom(32)
        encrypted = protocol.encrypt_message(None, session_key)
        decrypted, _ = protocol.decrypt_message(encrypted, session_key)
        assert decrypted is None

    def test_encrypt_nonce_is_random(self):
        """同じデータを暗号化しても毎回異なる暗号文になることを確認"""
        session_key = os.urandom(32)
        data = {"test": "value"}
        enc1 = protocol.encrypt_message(data, session_key)
        enc2 = protocol.encrypt_message(data, session_key)
        assert enc1 != enc2

    def test_encrypted_message_has_length_header(self):
        """暗号化メッセージが 4 バイトの長さヘッダーを持つことを確認"""
        import struct

        session_key = os.urandom(32)
        encrypted = protocol.encrypt_message("test", session_key)
        assert len(encrypted) >= 4
        length = struct.unpack("!I", encrypted[:4])[0]
        assert len(encrypted) == 4 + length

    def test_decrypt_wrong_key_returns_none(self):
        """異なる鍵で復号すると None が返されることを確認"""
        key1 = os.urandom(32)
        key2 = os.urandom(32)
        encrypted = protocol.encrypt_message({"test": 1}, key1)
        decrypted, _ = protocol.decrypt_message(encrypted, key2)
        assert decrypted is None

    def test_decrypt_tampered_ciphertext_returns_none(self):
        """改ざんされた暗号文の復号で None が返されることを確認"""
        session_key = os.urandom(32)
        encrypted = bytearray(protocol.encrypt_message({"test": 1}, session_key))
        encrypted[-1] ^= 0xFF  # 最後のバイトを反転
        decrypted, _ = protocol.decrypt_message(bytes(encrypted), session_key)
        assert decrypted is None

    def test_decrypt_truncated_data_returns_none(self):
        """途中で切れたデータの復号で None が返されることを確認"""
        session_key = os.urandom(32)
        encrypted = protocol.encrypt_message("hello", session_key)
        decrypted, _ = protocol.decrypt_message(encrypted[:10], session_key)
        assert decrypted is None

    def test_decrypt_empty_data_returns_none(self):
        """空データの復号で None が返されることを確認"""
        decrypted, rest = protocol.decrypt_message(b"", os.urandom(32))
        assert decrypted is None
        assert rest == b""

    def test_decrypt_too_short_body_returns_none(self):
        """最小サイズ (nonce+tag=28バイト) 未満の本体で None が返されることを確認"""
        import struct

        session_key = os.urandom(32)
        # 長さヘッダーが 10 を示すが本体は 10 バイト (28バイト未満)
        bad_data = struct.pack("!I", 10) + bytes(10)
        decrypted, _ = protocol.decrypt_message(bad_data, session_key)
        assert decrypted is None

    def test_encrypt_output_format(self):
        """暗号化出力のフォーマット確認: [4-byte length | 12-byte nonce | ciphertext+tag]"""
        import struct

        session_key = os.urandom(32)
        data = "test"
        encrypted = protocol.encrypt_message(data, session_key)

        # 4バイトのヘッダーを読む
        length = struct.unpack("!I", encrypted[:4])[0]
        body = encrypted[4:]
        assert len(body) == length
        # body の最初の 12 バイトが nonce
        assert length >= 12 + 16  # nonce + minimum GCM tag

    def test_original_decode_message_unaffected(self):
        """既存の decode_message() が引き続き動作することを確認 (後方互換性)"""
        data = {"key": "value", "number": 42}
        encoded = protocol.encode_message(data)
        decoded, rest = protocol.decode_message(encoded)
        assert decoded == data
        assert rest == b""



import datetime
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


def generate_certificate(cert_path="cert.pem", key_path="key.pem"):
    """QUIC/TLS 1.3用の自己署名証明書を生成する"""
    print(f"Generating certificate: {cert_path} and {key_path}...")

    # 秘密鍵の生成
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    # 証明書の作成
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "JP"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Tokyo"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "Chiyoda"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "NanaSQLite-Project"),
            x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
        ]
    )

    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
        .not_valid_after(
            datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=365)
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("localhost")]),
            critical=False,
        )
        .sign(private_key, hashes.SHA256())
    )

    # 鍵の保存
    with open(key_path, "wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    # 証明書の保存
    with open(cert_path, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))

    print("Certificate generated successfully.")


if __name__ == "__main__":
    generate_certificate()



import asyncio
import json
import os
import pytest
import ssl
from aioquic.asyncio import connect
from aioquic.quic.configuration import QuicConfiguration
from cryptography.hazmat.primitives import serialization
from nanasqlite_server.client import RemoteNanaSQLite, NanaRpcClientProtocol
from nanasqlite_server.cert_gen import generate_certificate
from nanasqlite_server.key_gen import generate_keys
import subprocess
import sys
import signal
import time


# テスト用のポート (conftest.py で設定される)
def get_port():
    return int(os.environ.get("NANASQLITE_TEST_PORT", 4433))


@pytest.fixture
def test_keys():
    """テスト用の Ed25519 鍵ペア"""
    from cryptography.hazmat.primitives.asymmetric import ed25519

    private_key = ed25519.Ed25519PrivateKey.generate()
    public_key = private_key.public_key()
    pub_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    ).decode()
    return private_key, pub_bytes


@pytest.fixture
async def dedicated_server(tmp_path):
    """各テスト用に独立したサーバーを起動するフィクスチャ"""
    # 鍵と証明書の準備
    priv_key_path = tmp_path / "nana_private.pem"
    config_path = tmp_path / "accounts.json"

    orig_cwd = os.getcwd()
    os.chdir(tmp_path)
    try:
        generate_certificate()
        generate_keys()

        with open(config_path, "w") as f:
            json.dump({"accounts": []}, f)

        # 未使用ポートの取得
        import socket

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        s.close()

        env = os.environ.copy()
        env["PYTHONPATH"] = orig_cwd + os.pathsep + os.path.join(orig_cwd, "src")
        env["PYTHONUNBUFFERED"] = "1"
        env["NANASQLITE_FORCE_POLLING"] = "1"

        db_dir = str(tmp_path)
        db_path = tmp_path / "dedicated_server_db.sqlite"
        
        # Provide db_dir in config
        with open(config_path, "w") as f:
            json.dump({"db_dir": db_dir, "accounts": []}, f)

        cmd = [
            sys.executable,
            "-m",
            "nanasqlite_server.server",
            "--port",
            str(port),
            "--accounts",
            str(config_path),
            # --db argument is deprecated/legacy but still accepted by argparse, 
            # effectively ignored by new logic if multi-DB is used correctly via accounts.
            # We keep it to avoid changing server argument parsing logic if not necessary.
            "--db", 
            str(db_path), 
        ]

        # パイプ詰まりによるハングアップを防ぐため
        log_file_path = tmp_path / "dedicated_server.log"
        log_file = open(log_file_path, "w", encoding="utf-8")

        kwargs = {}
        if sys.platform == "win32":
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

        proc = subprocess.Popen(
            cmd, env=env, stdout=log_file, stderr=subprocess.STDOUT, text=True, **kwargs
        )

        async def wait_for_quic():
            config = QuicConfiguration(is_client=True, verify_mode=ssl.CERT_NONE)
            start_wait = time.time()
            while time.time() - start_wait < 60.0:  # タイムアウトを延長
                if proc.poll() is not None:
                    return False
                try:
                    async with connect("127.0.0.1", port, configuration=config):
                        return True
                except Exception:
                    # Ignore connection errors during wait
                    await asyncio.sleep(1.0)
            return False

        if not await wait_for_quic():
            if proc.poll() is not None:
                log_file.close()
                with open(log_file_path, "r", encoding="utf-8") as f:
                    log_content = f.read()
                raise RuntimeError(
                    f"Dedicated server process died. Code: {proc.returncode}\nLog:\n{log_content}"
                )
            else:
                proc.kill()
                raise RuntimeError(
                    "Dedicated server failed to start (QUIC check failed)"
                )

        yield port, config_path, priv_key_path, db_dir

    finally:
        if proc.poll() is None:
            try:
                if sys.platform == "win32":
                    os.kill(proc.pid, signal.CTRL_BREAK_EVENT)
                else:
                    proc.send_signal(signal.SIGINT)
                try:
                    proc.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    proc.kill()
                    proc.wait()
            except Exception:
                proc.kill()
                proc.wait()
        log_file.close()
        os.chdir(orig_cwd)


@pytest.mark.asyncio
async def test_rbac_permissions(test_keys, dedicated_server):
    """RBAC: アカウントごとの権限制限が機能することを確認"""
    port, config_path, _, db_dir = dedicated_server
    priv, pub = test_keys

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "readonly",
                        "public_key": pub,
                        "allowed_methods": None,
                        "forbidden_methods": ["__setitem__"],
                        "allowed_dbs": ["dedicated_server_db.sqlite"]
                    }
                ]
            },
            f,
        )

    # 監視が反映されるまでCI環境では長めに待機
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv

    try:
        await client.connect(account_name="readonly")

        with pytest.raises(PermissionError) as excinfo:
            await client.set_item_async("key", "value", db="dedicated_server_db.sqlite")
        assert "forbidden" in str(excinfo.value).lower()

    finally:
        await client.close()


@pytest.mark.asyncio
async def test_realtime_policy_update(test_keys, dedicated_server):
    """即時反映: 実行中に権限を剥奪できることを確認"""
    port, config_path, _, db_dir = dedicated_server
    priv, pub = test_keys

    accounts = {
        "db_dir": db_dir,
        "accounts": [
            {
                "name": "user",
                "public_key": pub,
                "allowed_methods": None,
                "forbidden_methods": [],
                "allowed_dbs": ["dedicated_server_db.sqlite"]
            }
        ]
    }
    with open(config_path, "w") as f:
        json.dump(accounts, f)

    # 反映待ち
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv

    try:
        await client.connect(account_name="user")
        await client.list_tables(db="dedicated_server_db.sqlite")

        accounts["accounts"][0]["forbidden_methods"] = ["list_tables"]
        with open(config_path, "w") as f:
            json.dump(accounts, f)

        # watchfiles の検知待ち (CI環境では長めに)
        await asyncio.sleep(2.0)

        with pytest.raises(PermissionError):
            await client.list_tables(db="dedicated_server_db.sqlite")

    finally:
        await client.close()


@pytest.mark.asyncio
async def test_anti_dos_stream_limit(dedicated_server):
    """Anti-DoS: 未認証時の同時ストリーム数制限"""
    port, _, _, _ = dedicated_server
    config = QuicConfiguration(is_client=True, verify_mode=ssl.CERT_NONE)

    async with connect(
        "127.0.0.1", port, configuration=config, create_protocol=NanaRpcClientProtocol
    ) as conn:
        # Limit is 50. We send 60 to exceed it.
        for _ in range(60):
            try:
                stream_id = conn._quic.get_next_available_stream_id()
                conn._quic.send_stream_data(stream_id, b"data", end_stream=False)
                conn.transmit()
                # Give server time to process resets
                await asyncio.sleep(0.001)
            except Exception:
                # Connection might be reset or closed by server due to limits
                break

        await asyncio.sleep(1.0)
        async with connect(
            "127.0.0.1",
            port,
            configuration=config,
            create_protocol=NanaRpcClientProtocol,
        ) as conn2:
            # New connection should still be possible
            await conn2.call_rpc("AUTH_START")


@pytest.mark.asyncio
async def test_anti_dos_total_buffer_limit(dedicated_server):
    """Anti-DoS: 未認証時の合計バッファ制限"""
    port, _, _, _ = dedicated_server
    config = QuicConfiguration(is_client=True, verify_mode=ssl.CERT_NONE)

    async with connect(
        "127.0.0.1", port, configuration=config, create_protocol=NanaRpcClientProtocol
    ) as conn:
        stream_id = conn._quic.get_next_available_stream_id()
        chunk = b"A" * (1024 * 1024)  # 1MB

        # Limit is 50MB. We send 60MB.
        for _ in range(60):
            try:
                conn._quic.send_stream_data(stream_id, chunk, end_stream=False)
                conn.transmit()
                # Server needs time to process large data and close connection
                await asyncio.sleep(0.05)
            except Exception:
                # Connection closed by server due to buffer limits
                break

        await asyncio.sleep(1.0)
        try:
            # Original connection should be dead
            await conn.call_rpc("AUTH_START")
        except Exception:
            # Expected failure if connection is closed
            pass

        # New connection should still be possible
        async with connect(
            "127.0.0.1",
            port,
            configuration=config,
            create_protocol=NanaRpcClientProtocol,
        ) as conn2:
            await conn2.call_rpc("AUTH_START")



"""
NanaSQLite-Server v1.3.3 新機能テスト

1. read_only フラグ
2. DB名に紐づくテーブル名の動的制限
3. エラー情報の隠蔽強化
"""

import asyncio
import json
import os
import ssl
import sys
import signal
import subprocess
import time
import pytest

from aioquic.asyncio import connect
from aioquic.quic.configuration import QuicConfiguration
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from nanasqlite_server.accounts import Account, AccountManager
from nanasqlite_server.client import RemoteNanaSQLite
from nanasqlite_server.cert_gen import generate_certificate
from nanasqlite_server.key_gen import generate_keys


# ---------------------------------------------------------------------------
# ユニットテスト: Account / AccountManager
# ---------------------------------------------------------------------------


class TestAccountReadOnly:
    """read_only フラグの Account クラス単体テスト"""

    def test_read_only_default_false(self):
        acc = Account("user", "ssh-ed25519 AAAA fake_key")
        assert acc.read_only is False

    def test_read_only_true(self):
        acc = Account("viewer", "ssh-ed25519 AAAA fake_key", read_only=True)
        assert acc.read_only is True

    def test_read_only_false_explicit(self):
        acc = Account("admin", "ssh-ed25519 AAAA fake_key", read_only=False)
        assert acc.read_only is False


class TestAccountAllowedDbsFormat:
    """allowed_dbs の各フォーマット変換テスト"""

    def test_none_stays_none(self):
        acc = Account("u", "k", allowed_dbs=None)
        assert acc.allowed_dbs is None

    def test_list_converted_to_dict(self):
        acc = Account("u", "k", allowed_dbs=["a.sqlite", "b.sqlite"])
        assert isinstance(acc.allowed_dbs, dict)
        assert acc.allowed_dbs == {"a.sqlite": None, "b.sqlite": None}

    def test_dict_with_none_values(self):
        acc = Account("u", "k", allowed_dbs={"a.sqlite": None})
        assert acc.allowed_dbs == {"a.sqlite": None}

    def test_dict_with_table_list(self):
        acc = Account("u", "k", allowed_dbs={"a.sqlite": ["tbl1", "tbl2"]})
        assert acc.allowed_dbs == {"a.sqlite": {"tbl1", "tbl2"}}

    def test_dict_mixed(self):
        acc = Account(
            "u",
            "k",
            allowed_dbs={"logs.sqlite": None, "data.sqlite": ["public_info", "stats"]},
        )
        assert acc.allowed_dbs["logs.sqlite"] is None
        assert acc.allowed_dbs["data.sqlite"] == {"public_info", "stats"}

    def test_empty_list(self):
        acc = Account("u", "k", allowed_dbs=[])
        assert acc.allowed_dbs == {}


class TestAccountManagerLoadReadOnly:
    """AccountManager._do_load が read_only を正しく読み込むテスト"""

    def test_load_read_only_from_json(self, tmp_path):
        config = {
            "accounts": [
                {"name": "viewer", "public_key": "ssh-ed25519 AAAA fake", "read_only": True},
                {"name": "admin", "public_key": "ssh-ed25519 AAAA fake2"},
            ]
        }
        path = tmp_path / "accounts.json"
        path.write_text(json.dumps(config))
        mgr = AccountManager(str(path))
        assert mgr.accounts[0].read_only is True
        assert mgr.accounts[1].read_only is False

    def test_load_allowed_dbs_dict_from_json(self, tmp_path):
        config = {
            "accounts": [
                {
                    "name": "limited",
                    "public_key": "ssh-ed25519 AAAA fake",
                    "allowed_dbs": {"data.sqlite": ["t1", "t2"]},
                }
            ]
        }
        path = tmp_path / "accounts.json"
        path.write_text(json.dumps(config))
        mgr = AccountManager(str(path))
        acc = mgr.accounts[0]
        assert acc.allowed_dbs == {"data.sqlite": {"t1", "t2"}}

    def test_load_allowed_dbs_list_backward_compat(self, tmp_path):
        config = {
            "accounts": [
                {
                    "name": "user",
                    "public_key": "ssh-ed25519 AAAA fake",
                    "allowed_dbs": ["db1.sqlite", "db2.sqlite"],
                }
            ]
        }
        path = tmp_path / "accounts.json"
        path.write_text(json.dumps(config))
        mgr = AccountManager(str(path))
        acc = mgr.accounts[0]
        assert acc.allowed_dbs == {"db1.sqlite": None, "db2.sqlite": None}


# ---------------------------------------------------------------------------
# ユニットテスト: execute_rpc 権限チェック (サーバー内部ロジック)
# ---------------------------------------------------------------------------


class TestWriteMethodsConstant:
    """WRITE_METHODS 定数の検証"""

    def test_write_methods_defined(self):
        from nanasqlite_server.server import WRITE_METHODS
        assert "__setitem__" in WRITE_METHODS
        assert "__delitem__" in WRITE_METHODS
        assert "set" in WRITE_METHODS
        assert "delete" in WRITE_METHODS
        assert "batch_update" in WRITE_METHODS
        assert "clear" in WRITE_METHODS
        assert "upsert" in WRITE_METHODS

    def test_read_methods_not_in_write_methods(self):
        from nanasqlite_server.server import WRITE_METHODS
        assert "__getitem__" not in WRITE_METHODS
        assert "get" not in WRITE_METHODS
        assert "batch_get" not in WRITE_METHODS
        assert "list_tables" not in WRITE_METHODS
        assert "__contains__" not in WRITE_METHODS
        assert "__len__" not in WRITE_METHODS


# ---------------------------------------------------------------------------
# 統合テスト: 専用サーバーを使った read_only / テーブル制限テスト
# ---------------------------------------------------------------------------


@pytest.fixture(scope="module")
def test_keys_v133():
    """テスト用の Ed25519 鍵ペア (モジュール全体で1回生成)"""
    private_key = ed25519.Ed25519PrivateKey.generate()
    public_key = private_key.public_key()
    pub_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    ).decode()
    return private_key, pub_bytes


@pytest.fixture(scope="module")
def v133_server(tmp_path_factory, test_keys_v133):
    """v1.3.3 テスト用の専用サーバー (モジュール全体で1回起動)"""
    priv, pub = test_keys_v133
    tmp_path = tmp_path_factory.mktemp("v133")
    db_dir = str(tmp_path)
    config_path = tmp_path / "accounts.json"

    orig_cwd = os.getcwd()
    os.chdir(tmp_path)
    proc = None
    log_file = None
    try:
        generate_certificate()
        generate_keys()

        import socket

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        s.close()

        with open(config_path, "w") as f:
            json.dump({"db_dir": db_dir, "accounts": []}, f)

        env = os.environ.copy()
        env["PYTHONPATH"] = orig_cwd + os.pathsep + os.path.join(orig_cwd, "src")
        env["NANASQLITE_FORCE_POLLING"] = "1"
        env["NANASQLITE_DISABLE_BAN"] = "1"

        db_path = tmp_path / "v133_test.sqlite"
        cmd = [
            sys.executable,
            "-m",
            "nanasqlite_server.server",
            "--port",
            str(port),
            "--accounts",
            str(config_path),
            "--db",
            str(db_path),
        ]

        log_file_path = tmp_path / "v133_server.log"
        log_file = open(log_file_path, "w", encoding="utf-8")

        kwargs = {}
        if sys.platform == "win32":
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

        proc = subprocess.Popen(
            cmd, env=env, stdout=log_file, stderr=subprocess.STDOUT, text=True, **kwargs
        )

        # 同期的にサーバー起動を待つ (module-scoped fixture では asyncio を使えないため)
        quic_config = QuicConfiguration(is_client=True, verify_mode=ssl.CERT_NONE)
        start_wait = time.time()
        server_ready = False
        loop = asyncio.new_event_loop()
        try:
            while time.time() - start_wait < 60.0:
                if proc.poll() is not None:
                    break

                async def _check():
                    try:
                        async with connect("127.0.0.1", port, configuration=quic_config):
                            return True
                    except Exception:
                        return False

                result = loop.run_until_complete(_check())
                if result:
                    server_ready = True
                    break
                time.sleep(1.0)
        finally:
            loop.close()

        if not server_ready:
            log_file.close()
            with open(log_file_path, "r", encoding="utf-8") as f:
                content = f.read()
            raise RuntimeError(
                f"v133 server failed to start.\nLog:\n{content}"
            )

        yield port, config_path, priv, pub, db_dir

    finally:
        if proc and proc.poll() is None:
            try:
                if sys.platform == "win32":
                    os.kill(proc.pid, signal.CTRL_BREAK_EVENT)
                else:
                    proc.send_signal(signal.SIGINT)
                try:
                    proc.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    proc.kill()
                    proc.wait()
            except Exception:
                if proc:
                    proc.kill()
                    proc.wait()
        if log_file:
            log_file.close()
        os.chdir(orig_cwd)


@pytest.mark.asyncio
async def test_read_only_blocks_write_methods(v133_server):
    """read_only=true のアカウントは書き込み系メソッドを拒否される"""
    port, config_path, priv, pub, db_dir = v133_server

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "viewer",
                        "public_key": pub,
                        "read_only": True,
                        "allowed_dbs": ["v133_test.sqlite"],
                    }
                ],
            },
            f,
        )
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    try:
        await client.connect(account_name="viewer")

        # 書き込み系メソッドはすべて PermissionError になる
        with pytest.raises(PermissionError):
            await client.set_item_async("key", "value", db="v133_test.sqlite")
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_read_only_allows_read_methods(v133_server):
    """read_only=true のアカウントでも読み取り系メソッドは使用できる"""
    port, config_path, priv, pub, db_dir = v133_server

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "viewer",
                        "public_key": pub,
                        "read_only": True,
                        "allowed_dbs": ["v133_test.sqlite"],
                    }
                ],
            },
            f,
        )
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    try:
        await client.connect(account_name="viewer")
        # list_tables は読み取り系なので PermissionError にならない
        result = await client.list_tables(db="v133_test.sqlite")
        assert result is not None
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_table_restriction_blocks_unauthorized_table(v133_server):
    """allowed_dbs に dict 形式でテーブル制限を設定すると、未許可テーブルへのアクセスが拒否される"""
    port, config_path, priv, pub, db_dir = v133_server

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "limited",
                        "public_key": pub,
                        "allowed_dbs": {
                            "v133_test.sqlite": ["allowed_table"],
                        },
                    }
                ],
            },
            f,
        )
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    try:
        await client.connect(account_name="limited")

        # 許可されていないテーブルへのアクセスは PermissionError
        # __getattr__ を使って table_name を kwargs に直接渡す
        with pytest.raises(PermissionError):
            await client.__getattr__("set")(
                "key", "value", db="v133_test.sqlite", table_name="secret_table"
            )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_table_restriction_null_allows_all_tables(v133_server):
    """allowed_dbs に null を設定すると全テーブルへのアクセスが許可される"""
    port, config_path, priv, pub, db_dir = v133_server

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "full_access",
                        "public_key": pub,
                        "allowed_dbs": {
                            "v133_test.sqlite": None,
                        },
                    }
                ],
            },
            f,
        )
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    try:
        await client.connect(account_name="full_access")
        # テーブル制限なし: 書き込みも正常に動作する
        await client.set_item_async("key", "value", db="v133_test.sqlite")
        result = await client.get_item_async("key", db="v133_test.sqlite")
        assert result == "value"
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_backward_compat_list_allowed_dbs(v133_server):
    """allowed_dbs にリスト形式を使用しても引き続き動作する (後方互換)"""
    port, config_path, priv, pub, db_dir = v133_server

    with open(config_path, "w") as f:
        json.dump(
            {
                "db_dir": db_dir,
                "accounts": [
                    {
                        "name": "compat_user",
                        "public_key": pub,
                        "allowed_dbs": ["v133_test.sqlite"],
                    }
                ],
            },
            f,
        )
    await asyncio.sleep(2.0)

    client = RemoteNanaSQLite(host="127.0.0.1", port=port, verify_ssl=False)
    client.private_key = priv
    try:
        await client.connect(account_name="compat_user")
        await client.set_item_async("compat_key", "compat_value", db="v133_test.sqlite")
        result = await client.get_item_async("compat_key", db="v133_test.sqlite")
        assert result == "compat_value"
    finally:
        await client.close()


# ---------------------------------------------------------------------------
# ユニットテスト: エラー情報隠蔽 (AttributeError/TypeError → InternalServerError)
# ---------------------------------------------------------------------------


class TestErrorHiding:
    """handle_request での内部エラー隠蔽テスト"""

    def _make_protocol(self):
        """テスト用の NanaRpcProtocol インスタンスを作成する"""
        from unittest.mock import MagicMock
        from nanasqlite_server.server import NanaRpcProtocol

        account_manager = MagicMock()
        quic = MagicMock()
        quic.get_next_available_stream_id.return_value = 0
        protocol = NanaRpcProtocol.__new__(NanaRpcProtocol)
        protocol.authenticated = True
        protocol.account = MagicMock()
        protocol.account.name = "test_user"
        protocol.account_manager = account_manager
        protocol.last_db_name = "test.sqlite"
        protocol.pqc_session_key = None
        protocol._quic = quic
        protocol._background_tasks = set()
        protocol._kem_instance = None
        protocol.client_ip = "127.0.0.1"
        return protocol

    def _make_rpc_payload(self):
        """有効な RPC リクエストのエンコード済みバイト列を返す"""
        from nanasqlite_server import protocol as proto
        return proto.encode_message({"method": "list_tables", "args": [], "kwargs": {}})

    @pytest.mark.asyncio
    async def test_attribute_error_returns_internal_server_error(self):
        """AttributeError は InternalServerError として返され、詳細は露出しない"""
        from unittest.mock import patch, AsyncMock

        protocol = self._make_protocol()

        sent_responses = []

        def capture_send(stream_id, data):
            sent_responses.append(data)

        protocol._send_response = capture_send

        with patch.object(
            protocol,
            "execute_rpc",
            new=AsyncMock(side_effect=AttributeError("secret internal detail")),
        ):
            await protocol.handle_request(0, self._make_rpc_payload())

        assert len(sent_responses) == 1
        resp = sent_responses[0]
        assert isinstance(resp, dict)
        assert resp.get("error_type") == "InternalServerError"
        assert "secret internal detail" not in resp.get("message", "")
        assert resp.get("message") == "Internal Server Error"

    @pytest.mark.asyncio
    async def test_type_error_returns_internal_server_error(self):
        """TypeError は InternalServerError として返され、詳細は露出しない"""
        from unittest.mock import patch, AsyncMock

        protocol = self._make_protocol()

        sent_responses = []

        def capture_send(stream_id, data):
            sent_responses.append(data)

        protocol._send_response = capture_send

        with patch.object(
            protocol,
            "execute_rpc",
            new=AsyncMock(side_effect=TypeError("unexpected type info")),
        ):
            await protocol.handle_request(0, self._make_rpc_payload())

        assert len(sent_responses) == 1
        resp = sent_responses[0]
        assert isinstance(resp, dict)
        assert resp.get("error_type") == "InternalServerError"
        assert "unexpected type info" not in resp.get("message", "")
        assert resp.get("message") == "Internal Server Error"

    @pytest.mark.asyncio
    async def test_permission_error_message_is_returned_to_client(self):
        """PermissionError のメッセージはそのままクライアントに返される"""
        from unittest.mock import patch, AsyncMock

        protocol = self._make_protocol()

        sent_responses = []

        def capture_send(stream_id, data):
            sent_responses.append(data)

        protocol._send_response = capture_send

        with patch.object(
            protocol,
            "execute_rpc",
            new=AsyncMock(side_effect=PermissionError("access denied for user")),
        ):
            await protocol.handle_request(0, self._make_rpc_payload())

        assert len(sent_responses) == 1
        resp = sent_responses[0]
        assert isinstance(resp, dict)
        assert resp.get("error_type") == "PermissionError"
        assert "access denied for user" in resp.get("message", "")



"""
PQC (Post-Quantum Cryptography) サポートのテスト

liboqs-python が利用できる場合に PQC 鍵の生成、アカウント登録、
署名検証、KEM鍵交換、暗号化通信が正しく動作することを検証します。
liboqs-python が不要なテスト (test_generate_pqc_keys_no_oqs_raises 等) は
liboqs-python がない環境でも実行されます。
"""

import base64
import json
import os
import secrets
import pytest

# liboqs が利用できるか確認
try:
    import oqs
    HAS_OQS = True
except (ImportError, SystemExit):
    oqs = None
    HAS_OQS = False

_skip_no_oqs = pytest.mark.skipif(
    not HAS_OQS,
    reason="liboqs-python is not installed. Install with: pip install liboqs-python",
)


@pytest.fixture(scope="session")
def pqc_keypair():
    """Dilithium3 鍵ペアを生成して返す (セッション内で再利用)"""
    with oqs.Signature("Dilithium3") as signer:
        public_key_bytes = signer.generate_keypair()
        secret_key_bytes = signer.export_secret_key()
    return {
        "algorithm": "Dilithium3",
        "public_key_bytes": public_key_bytes,
        "secret_key_bytes": secret_key_bytes,
        "public_key_str": f"oqs-Dilithium3:{base64.b64encode(public_key_bytes).decode()}",
    }


class TestPqcKeyGeneration:
    """PQC 鍵生成のテスト"""

    @_skip_no_oqs
    def test_generate_pqc_keys_creates_files(self, tmp_path, monkeypatch):
        """generate_pqc_keys() が正しくファイルを作成することを確認"""
        monkeypatch.chdir(tmp_path)
        from nanasqlite_server.key_gen import generate_pqc_keys

        generate_pqc_keys(algorithm="Dilithium3", prefix="test")

        private_path = tmp_path / "test_private_pqc.json"
        public_path = tmp_path / "test_public_pqc.pub"

        assert private_path.exists(), "Private key file should be created"
        assert public_path.exists(), "Public key file should be created"

    @_skip_no_oqs
    def test_private_key_file_format(self, tmp_path, monkeypatch):
        """秘密鍵ファイルが正しい JSON 形式であることを確認"""
        monkeypatch.chdir(tmp_path)
        from nanasqlite_server.key_gen import generate_pqc_keys

        generate_pqc_keys(algorithm="Dilithium3", prefix="test")

        with open(tmp_path / "test_private_pqc.json") as f:
            data = json.load(f)

        assert data["algorithm"] == "Dilithium3"
        assert "secret_key" in data
        # Base64 デコードできることを確認
        decoded = base64.b64decode(data["secret_key"])
        assert len(decoded) > 0

    @_skip_no_oqs
    def test_public_key_file_format(self, tmp_path, monkeypatch):
        """公開鍵ファイルが正しい形式であることを確認"""
        monkeypatch.chdir(tmp_path)
        from nanasqlite_server.key_gen import generate_pqc_keys

        generate_pqc_keys(algorithm="Dilithium3", prefix="test")

        with open(tmp_path / "test_public_pqc.pub") as f:
            content = f.read().strip()

        assert content.startswith("oqs-Dilithium3:"), (
            f"Public key should start with 'oqs-Dilithium3:', got: {content[:50]}"
        )
        # Base64 部分がデコードできることを確認
        b64_part = content.split(":", 1)[1]
        decoded = base64.b64decode(b64_part)
        assert len(decoded) > 0

    def test_generate_pqc_keys_no_oqs_raises(self, monkeypatch):
        """liboqs-python がない場合に ImportError が発生することを確認"""
        import sys
        from nanasqlite_server.key_gen import generate_pqc_keys

        # sys.modules 経由でモジュール参照を取得し HAS_OQS を一時的に False に設定
        kg_module = sys.modules.get("nanasqlite_server.key_gen")
        assert kg_module is not None, "nanasqlite_server.key_gen should be imported"
        monkeypatch.setattr(kg_module, "HAS_OQS", False)
        with pytest.raises(ImportError, match="liboqs-python"):
            generate_pqc_keys()


@_skip_no_oqs
class TestPqcAccount:
    """PQC アカウントの作成と署名検証のテスト"""

    def test_account_detects_pqc_key(self, pqc_keypair):
        """Account クラスが PQC 公開鍵を正しく検出することを確認"""
        from nanasqlite_server.accounts import Account

        account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        assert account.key_type == "oqs"
        assert account.pqc_algorithm == "Dilithium3"
        assert account.pqc_public_key_bytes is not None
        assert account.public_key is None  # Ed25519 key should not be set

    def test_account_detects_ed25519_key(self):
        """Account クラスが Ed25519 公開鍵を正しく検出することを確認"""
        from cryptography.hazmat.primitives.asymmetric import ed25519
        from cryptography.hazmat.primitives import serialization
        from nanasqlite_server.accounts import Account

        priv = ed25519.Ed25519PrivateKey.generate()
        pub = priv.public_key()
        pub_bytes = pub.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
        )

        account = Account(name="ed_user", public_key_pem=pub_bytes.decode())

        assert account.key_type == "ed25519"
        assert account.pqc_algorithm is None
        assert account.pqc_public_key_bytes is None
        assert account.public_key is not None

    def test_pqc_signature_verification(self, pqc_keypair):
        """PQC 署名の検証が正しく動作することを確認"""
        from nanasqlite_server.accounts import Account, AccountManager

        challenge = secrets.token_bytes(32)

        # 署名
        with oqs.Signature(
            pqc_keypair["algorithm"], pqc_keypair["secret_key_bytes"]
        ) as signer:
            signature = signer.sign(challenge)

        account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        # _verify_account_signature をテスト
        manager = object.__new__(AccountManager)
        result = manager._verify_account_signature(account, signature, challenge)
        assert result is True

    def test_pqc_wrong_signature_rejected(self, pqc_keypair):
        """誤った PQC 署名が拒否されることを確認"""
        from nanasqlite_server.accounts import Account, AccountManager

        challenge = secrets.token_bytes(32)
        wrong_signature = secrets.token_bytes(64)  # ランダムバイト = 不正な署名

        account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        manager = object.__new__(AccountManager)
        result = manager._verify_account_signature(account, wrong_signature, challenge)
        assert result is False

    def test_find_account_by_pqc_signature(self, pqc_keypair):
        """AccountManager が PQC 署名でアカウントを検索できることを確認"""
        from nanasqlite_server.accounts import Account, AccountManager

        challenge = secrets.token_bytes(32)

        # 署名
        with oqs.Signature(
            pqc_keypair["algorithm"], pqc_keypair["secret_key_bytes"]
        ) as signer:
            signature = signer.sign(challenge)

        account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        manager = object.__new__(AccountManager)
        manager.accounts = [account]

        found = manager.find_account_by_signature(signature, challenge, "pqc_user")
        assert found is not None
        assert found.name == "pqc_user"

    def test_find_account_pqc_wrong_signature_returns_none(self, pqc_keypair):
        """誤った署名では None が返されることを確認"""
        from nanasqlite_server.accounts import Account, AccountManager

        challenge = secrets.token_bytes(32)
        wrong_signature = secrets.token_bytes(64)

        account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        manager = object.__new__(AccountManager)
        manager.accounts = [account]

        found = manager.find_account_by_signature(wrong_signature, challenge)
        assert found is None


@_skip_no_oqs
class TestPqcKeyFormat:
    """PQC 公開鍵フォーマットのテスト"""

    def test_invalid_pqc_key_format_handled_gracefully(self):
        """不正な PQC 鍵フォーマットが適切に処理されることを確認"""
        from nanasqlite_server.accounts import Account

        account = Account(
            name="bad_pqc_user",
            public_key_pem="oqs-Dilithium3:NOT_VALID_BASE64!!!",
        )

        # エラーで例外が出ずに、属性が None になっていることを確認
        assert account.key_type == "oqs"
        assert account.pqc_public_key_bytes is None

    def test_mixed_accounts_verification(self, pqc_keypair):
        """Ed25519 と PQC アカウントが混在する場合の検索をテスト"""
        from cryptography.hazmat.primitives.asymmetric import ed25519
        from cryptography.hazmat.primitives import serialization
        from nanasqlite_server.accounts import Account, AccountManager

        # Ed25519 アカウント
        priv = ed25519.Ed25519PrivateKey.generate()
        pub = priv.public_key()
        pub_bytes = pub.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
        )
        ed_account = Account(name="ed_user", public_key_pem=pub_bytes.decode())

        # PQC アカウント
        pqc_account = Account(
            name="pqc_user",
            public_key_pem=pqc_keypair["public_key_str"],
        )

        manager = object.__new__(AccountManager)
        manager.accounts = [ed_account, pqc_account]

        challenge = secrets.token_bytes(32)

        # PQC 署名で PQC アカウントが見つかることを確認
        with oqs.Signature(
            pqc_keypair["algorithm"], pqc_keypair["secret_key_bytes"]
        ) as signer:
            pqc_sig = signer.sign(challenge)

        found = manager.find_account_by_signature(pqc_sig, challenge)
        assert found is not None
        assert found.name == "pqc_user"

        # Ed25519 署名で Ed25519 アカウントが見つかることを確認
        ed_sig = priv.sign(challenge)
        found_ed = manager.find_account_by_signature(ed_sig, challenge)
        assert found_ed is not None
        assert found_ed.name == "ed_user"


class TestProtocolEncryptDecrypt:
    """AES-256-GCM 暗号化/復号のテスト (liboqs 不要)"""

    # このクラスのテストは liboqs なしでも動作する

    def test_derive_session_key_deterministic(self):
        """同じ共有秘密から常に同じセッション鍵が導出されることを確認"""
        from nanasqlite_server import protocol

        shared = b"test_shared_secret_32_bytes_here!"
        key1 = protocol.derive_session_key(shared)
        key2 = protocol.derive_session_key(shared)
        assert key1 == key2
        assert len(key1) == 32  # 256-bit

    def test_derive_session_key_different_inputs(self):
        """異なる共有秘密から異なるセッション鍵が導出されることを確認"""
        from nanasqlite_server import protocol

        key1 = protocol.derive_session_key(b"secret_a")
        key2 = protocol.derive_session_key(b"secret_b")
        assert key1 != key2

    def test_encrypt_decrypt_roundtrip_dict(self):
        """辞書データの暗号化→復号が正しく動作することを確認"""
        from nanasqlite_server import protocol

        session_key = os.urandom(32)
        data = {"method": "get_item_async", "args": ["key"], "kwargs": {}}
        encrypted = protocol.encrypt_message(data, session_key)
        decrypted, rest = protocol.decrypt_message(encrypted, session_key)
        assert decrypted == data
        assert rest == b""

    def test_encrypt_decrypt_roundtrip_string(self):
        """文字列の暗号化→復号が正しく動作することを確認"""
        from nanasqlite_server import protocol

        session_key = os.urandom(32)
        data = "AUTH_OK"
        encrypted = protocol.encrypt_message(data, session_key)
        decrypted, rest = protocol.decrypt_message(encrypted, session_key)
        assert decrypted == data
        assert rest == b""

    def test_encrypt_nonce_is_random(self):
        """同じデータの暗号化でも毎回異なる暗号文が生成されることを確認"""
        from nanasqlite_server import protocol

        session_key = os.urandom(32)
        data = {"test": "value"}
        enc1 = protocol.encrypt_message(data, session_key)
        enc2 = protocol.encrypt_message(data, session_key)
        assert enc1 != enc2  # nonce がランダムのため毎回異なる

    def test_decrypt_wrong_key_returns_none(self):
        """異なるセッション鍵で復号した場合 None が返されることを確認"""
        from nanasqlite_server import protocol

        key1 = os.urandom(32)
        key2 = os.urandom(32)
        encrypted = protocol.encrypt_message({"test": 1}, key1)
        decrypted, _ = protocol.decrypt_message(encrypted, key2)
        assert decrypted is None

    def test_decrypt_tampered_ciphertext_returns_none(self):
        """改ざんされた暗号文の復号で None が返されることを確認 (AEADの整合性検証)"""
        from nanasqlite_server import protocol

        session_key = os.urandom(32)
        encrypted = bytearray(protocol.encrypt_message({"test": 1}, session_key))
        # 最後のバイトを改ざん
        encrypted[-1] ^= 0xFF
        decrypted, _ = protocol.decrypt_message(bytes(encrypted), session_key)
        assert decrypted is None

    def test_decrypt_truncated_data_returns_none(self):
        """データが途中で切れた場合に None が返されることを確認"""
        from nanasqlite_server import protocol

        session_key = os.urandom(32)
        encrypted = protocol.encrypt_message("hello", session_key)
        decrypted, _ = protocol.decrypt_message(encrypted[:10], session_key)
        assert decrypted is None

    def test_decrypt_empty_data_returns_none(self):
        """空データの復号で None が返されることを確認"""
        from nanasqlite_server import protocol

        decrypted, rest = protocol.decrypt_message(b"", os.urandom(32))
        assert decrypted is None


@_skip_no_oqs
class TestPqcKemExchange:
    """PQC KEM 鍵交換のテスト (liboqs 必須)"""

    def test_kem_kyber768_shared_secret_matches(self):
        """Kyber768 KEM でクライアントとサーバーの共有秘密が一致することを確認"""
        with oqs.KeyEncapsulation("Kyber768") as kem_server:
            server_pubkey = kem_server.generate_keypair()
            with oqs.KeyEncapsulation("Kyber768") as kem_client:
                ciphertext, client_ss = kem_client.encap_secret(server_pubkey)
            server_ss = kem_server.decap_secret(ciphertext)
        assert client_ss == server_ss

    def test_kem_derived_session_keys_match(self):
        """KEM で交換した共有秘密から導出したセッション鍵が一致することを確認"""
        from nanasqlite_server import protocol

        with oqs.KeyEncapsulation("Kyber768") as kem_server:
            server_pubkey = kem_server.generate_keypair()
            with oqs.KeyEncapsulation("Kyber768") as kem_client:
                ciphertext, client_ss = kem_client.encap_secret(server_pubkey)
            server_ss = kem_server.decap_secret(ciphertext)

        client_key = protocol.derive_session_key(client_ss)
        server_key = protocol.derive_session_key(server_ss)
        assert client_key == server_key
        assert len(client_key) == 32

    def test_kem_encrypted_message_roundtrip(self):
        """KEM で交換したセッション鍵で暗号化→復号が正しく動作することを確認"""
        from nanasqlite_server import protocol

        with oqs.KeyEncapsulation("Kyber768") as kem_server:
            server_pubkey = kem_server.generate_keypair()
            with oqs.KeyEncapsulation("Kyber768") as kem_client:
                ciphertext, client_ss = kem_client.encap_secret(server_pubkey)
            kem_server.decap_secret(ciphertext)

        session_key = protocol.derive_session_key(client_ss)
        data = {"method": "get_item_async", "args": ["test_key"], "kwargs": {}}

        # サーバーが暗号化、クライアントが復号 (同じセッション鍵)
        encrypted = protocol.encrypt_message(data, session_key)
        decrypted, _ = protocol.decrypt_message(encrypted, session_key)
        assert decrypted == data

    def test_kem_different_connections_have_different_keys(self):
        """接続ごとに異なるKEM鍵ペアが生成されることを確認 (前方秘匿性)"""
        from nanasqlite_server import protocol

        session_keys = set()
        for _ in range(3):
            with oqs.KeyEncapsulation("Kyber768") as kem_server:
                server_pubkey = kem_server.generate_keypair()
                with oqs.KeyEncapsulation("Kyber768") as kem_client:
                    ciphertext, client_ss = kem_client.encap_secret(server_pubkey)
                server_ss = kem_server.decap_secret(ciphertext)
            session_key = protocol.derive_session_key(server_ss)
            session_keys.add(session_key)

        # 3回全て異なるセッション鍵が生成されるべき
        assert len(session_keys) == 3

    def test_kem_free_releases_instance(self):
        """KEM インスタンスの free() が正しく動作することを確認"""
        kem = oqs.KeyEncapsulation("Kyber768")
        kem.generate_keypair()
        # free() がエラーなく呼べることを確認
        kem.free()
